/* DIV.C - Tecnica da subdivisao dos dominios */

const float epsilon = 1e-6;

#define DET(a, b, c, d, e, f, g, h, i)  ((a)*(e)*(i)+(c)*(d)*(h)+(b)*(f)*(g)\
				       -(c)*(e)*(g)-(a)*(f)*(h)-(b)*(d)*(i))
#define SQR(x)  ((x)*(x))
#define DISTANCIA(a, b, c, x, y, z) sqrt(SQR(a-x) + SQR(b-y) + SQR(c-z))
#define PROD_VET(V1, V2, V3, a1, a2, a3, b1, b2, b3) V1 = a2*b3-a3*b2; \
				   V2 = a3*b1-a1*b3; V3 = a1*b2-a2*b1
#define NORMA(V1, V2, V3)  (sqrt((V1)*(V1) + (V2)*(V2) + (V3)*(V3)))

/* ------------------------------------------------------------------------- */

int VetorTangente(float *t1, float *t2, float *t3, float (*f1)(float, float),
		float (*f2)(float, float), float (*f3)(float, float),
		float u0, float v0, int direcao) {

  const float h = 1E-6;
  float norma;
  const float epsilon = 1e-6;

  if (direcao == 1) {
    *t1 = (f1(u0 + h, v0) - f1(u0 - h, v0))/(2.0*h);
    *t2 = (f2(u0 + h, v0) - f2(u0 - h, v0))/(2.0*h);
    *t3 = (f3(u0 + h, v0) - f3(u0 - h, v0))/(2.0*h);
  } else {
    *t1 = (f1(u0, v0 + h) - f1(u0, v0 - h))/(2.0*h);
    *t2 = (f2(u0, v0 + h) - f2(u0, v0 - h))/(2.0*h);
    *t3 = (f3(u0, v0 + h) - f3(u0, v0 - h))/(2.0*h);
  }

  norma = NORMA(*t1, *t2, *t3);
  if (norma < epsilon) return 0;
  return 1;
}

/* ------------------------------------------------------------------------ */

int VetorNormal(float *N0, float *N1, float *N2, float (*f1)(float, float),
		float (*f2)(float, float), float (*f3)(float, float),
		float u0, float v0) {

/* Calcula (*N0, *N1, *N2) o vetor normal no  ponto  (u0, v0)  `a
   superficie f(u, v) = (f1(u, v), f2(u, v), f3(u, v)). Retorna 0
   se o vetor normal encontrado for nulo ou 1 em  caso contrario. */

  const float h = 1E-6;
  float a1, a2, a3, b1, b2, b3, norma_N;
  const float epsilon = 1e-6;

  /* Calculo de 2 vetores tangentes `a superficie f */

  a1 = (f1(u0 + h, v0) - f1(u0 - h, v0))/(2.0*h);
  a2 = (f2(u0 + h, v0) - f2(u0 - h, v0))/(2.0*h);
  a3 = (f3(u0 + h, v0) - f3(u0 - h, v0))/(2.0*h);

  b1 = (f1(u0, v0 + h) - f1(u0, v0 - h))/(2.0*h);
  b2 = (f2(u0, v0 + h) - f2(u0, v0 - h))/(2.0*h);
  b3 = (f3(u0, v0 + h) - f3(u0, v0 - h))/(2.0*h);

  /* Vetor normal a f */
  PROD_VET(*N0, *N1, *N2, a1, a2, a3, b1, b2, b3);

  norma_N = NORMA(*N0, *N1, *N2);
  if (norma_N < epsilon) return 0;
  return 1;
}

/* ----------------------------------------------------------------------- */

int InterceptaSegmentos(float a1, float a2, float a3, float b1, float b2, float b3,
	       float c1, float c2, float c3, float d1, float d2, float d3,
	       float *solucao) {

/* Verifica se o segmento (a1, a2, a3) - (b1, b2, b3) intercepta em um
   unico ponto o segmento (c1, c2, c3) - (d1, d2, d3). E' resolvido  o
   sistema linear

		   t (b1 - a1) + s (c1 - d1) = c1 - a1
		   t (b2 - a2) + s (c2 - d2) = c2 - a2
		   t (b3 - a3) + s (c3 - d3) = c3 - a3

 */

  float t, s, determinante;

  determinante = (b1 - a1)*(c2 - d2) - (b2 - a2)*(c1 - d1);
  if (fabs(determinante) > epsilon) {
    t = ((c1 - a1)*(c2 - d2) - (c2 - a2)*(c1 - d1))/determinante;
    s = ((b1 - a1)*(c2 - a2) - (b2 - a2)*(c1 - a1))/determinante;
    *solucao = t;
    if (fabs(t*(b3 - a3) + s*(c3 - d3) - c3 + a3) < epsilon)
      return (-epsilon <= s && s <= 1 + epsilon);
    else
      return 0;
  }

  determinante = (b1 - a1)*(c3 - d3) - (b3 - a3)*(c1 - d1);
  if (fabs(determinante) > epsilon) {
    t = ((c1 - a1)*(c3 - d3) - (c3 - a3)*(c1 - d1))/determinante;
    s = ((b1 - a1)*(c3 - a3) - (b3 - a3)*(c1 - a1))/determinante;
    *solucao = t;
    if (fabs(t*(b2 - a2) + s*(c2 - d2) - c2 + a2) < epsilon)
      return (-epsilon <= s && s <= 1 + epsilon);
    else
      return 0;
  }

  determinante = (b2 - a2)*(c3 - d3) - (b3 - a3)*(c2 - d2);
  if (fabs(determinante) > epsilon) {
    t = ((c2 - a2)*(c3 - d3) - (c3 - a3)*(c2 - d2))/determinante;
    s = ((b2 - a2)*(c3 - a3) - (b3 - a3)*(c2 - a2))/determinante;
    *solucao = t;
    if (fabs(t*(b1 - a1) + s*(c1 - d1) - c1 + a1) < epsilon)
      return (-epsilon <= s && s <= 1 + epsilon);
    else
      return 0;
  }

  return 0;
}

/* ----------------------------------------------------------------------- */
/*
int InterceptaTriangulos(float p1_1, float p1_2, float p1_3,
			 float p2_1, float p2_2, float p2_3,
			 float p3_1, float p3_2, float p3_3,
			 float q1_1, float q1_2, float q1_3,
			 float q2_1, float q2_2, float q2_3,
			 float q3_1, float q3_2, float q3_3,
			 float *r1_1, float *r1_2, float *r1_3,
			 float *r2_1, float *r2_2, float *r2_3) {

  int teste1, teste2, teste3, quant = 0;
  float t1, t2, t3;

  teste1 = InterceptaSegmentos(p1_1, p1_2, p1_3, p2_1, p2_2, p2_3,
		      q1_1, q1_2, q1_3, q2_1, q2_2, q2_3, &t1);
  if (teste1) {
    *r1_1 = p1_1 + t1*(p2_1 - p1_1);
    *r1_2 = p1_2 + t1*(p2_2 - p1_2);
    *r1_3 = p1_3 + t1*(p2_3 - p1_3);
    quant++;
  }

  teste2 = InterceptaSegmentos(p2_1, p2_2, p2_3, p3_1, p3_2, p3_3,
		      q2_1, q2_2, q2_3, q3_1, q3_2, q3_3, &t2);
  if (teste2) {
    if (quant == 0) {
      *r1_1 = p2_1 + t2*(p3_1 - p2_1);
      *r1_2 = p2_2 + t2*(p3_2 - p2_2);
      *r1_3 = p2_3 + t2*(p3_3 - p2_3);
    } else {
      *r2_1 = p2_1 + t2*(p3_1 - p2_1);
      *r2_2 = p2_2 + t2*(p3_2 - p2_2);
      *r2_3 = p2_3 + t2*(p3_3 - p2_3);
      return 1;
    }
  }

  teste3 = InterceptaSegmentos(p3_1, p3_2, p3_3, p1_1, p1_2, p1_3,
		      q3_1, q3_2, q3_3, q1_1, q1_2, q1_3, &t3);
  if (teste3) {
    if (quant == 0) {
      *r1_1 = p3_1 + t3*(p1_1 - p3_1);
      *r1_2 = p3_2 + t3*(p1_2 - p3_2);
      *r1_3 = p3_3 + t3*(p1_3 - p3_3);
    } else {
      *r2_1 = p3_1 + t3*(p1_1 - p3_1);
      *r2_2 = p3_2 + t3*(p1_2 - p3_2);
      *r2_3 = p3_3 + t3*(p1_3 - p3_3);
      return 1;
    }
  }
  return 0;
}               */

/* ----------------------------------------------------------------------- */

float Angulo(float (*f1)(float, float), float (*f2)(float, float),
      float (*f3)(float, float), float u1, float u2, float v1, float v2) {

  float umed, vmed,
	nmedx, nmedy, nmedz,
	n1x, n1y, n1z,
	n2x, n2y, n2z,
	n3x, n3y, n3z,
	n4x, n4y, n4z, ang, ang_max = -1000;

  umed = (u1 + u2)/2.0, vmed = (v1 + v2)/2.0;

  VetorNormal(&nmedx, &nmedy, &nmedz, f1, f2, f3, umed, vmed);
  VetorNormal(&n1x, &n1y, &n1z, f1, f2, f3, u1, v1);
  VetorNormal(&n2x, &n2y, &n2z, f1, f2, f3, u1, v2);
  VetorNormal(&n3x, &n3y, &n3z, f1, f2, f3, u2, v1);
  VetorNormal(&n4x, &n4y, &n4z, f1, f2, f3, u2, v2);
  ang = acos((nmedx*n1x + nmedy*n1y + nmedz*n1z)/
	(NORMA(nmedx, nmedy, nmedz)*NORMA(n1x, n1y, n1z)));
  if (ang_max < ang) ang_max = ang;
  ang = acos((nmedx*n2x + nmedy*n2y + nmedz*n2z)/
	(NORMA(nmedx, nmedy, nmedz)*NORMA(n2x, n2y, n2z)));
  if (ang_max < ang) ang_max = ang;
  ang = acos((nmedx*n3x + nmedy*n3y + nmedz*n3z)/
	(NORMA(nmedx, nmedy, nmedz)*NORMA(n3x, n3y, n3z)));
  if (ang_max < ang) ang_max = ang;
  ang = acos((nmedx*n4x + nmedy*n4y + nmedz*n4z)/
	(NORMA(nmedx, nmedy, nmedz)*NORMA(n4x, n4y, n4z)));
  if (ang_max < ang) ang_max = ang;
  return ang_max;
}

/* ------------------------------------------------------------------------- */

float Bordas(float (*f1)(float, float), float (*f2)(float, float),
      float (*f3)(float, float), float u1, float u2, float v1, float v2) {

  float umed, vmed,
	v1x, v1y, v1z,
	v2x, v2y, v2z,
	v3x, v3y, v3z,
	ang, ang_max = -1000;

  umed = (u1 + u2)/2.0, vmed = (v1 + v2)/2.0;

  VetorTangente(&v1x, &v1y, &v1z, f1, f2, f3, u1, v1, 1);
  VetorTangente(&v2x, &v2y, &v2z, f1, f2, f3, u1, vmed, 1);
  VetorTangente(&v3x, &v3y, &v3z, f1, f2, f3, u1, v2, 1);
  ang = acos((v1x*v2x + v1y*v2y + v1z*v2z)/
	(NORMA(v1x, v1y, v1z)*NORMA(v2x, v2y, v2z)));
  if (ang_max < ang) ang_max = ang;
  ang = acos((v2x*v3x + v2y*v3y + v2z*v3z)/
	(NORMA(v2x, v2y, v2z)*NORMA(v3x, v3y, v3z)));
  if (ang_max < ang) ang_max = ang;

  VetorTangente(&v1x, &v1y, &v1z, f1, f2, f3, u2, v1, 1);
  VetorTangente(&v2x, &v2y, &v2z, f1, f2, f3, u2, vmed, 1);
  VetorTangente(&v3x, &v3y, &v3z, f1, f2, f3, u2, v2, 1);
  ang = acos((v1x*v2x + v1y*v2y + v1z*v2z)/
	(NORMA(v1x, v1y, v1z)*NORMA(v2x, v2y, v2z)));
  if (ang_max < ang) ang_max = ang;
  ang = acos((v2x*v3x + v2y*v3y + v2z*v3z)/
	(NORMA(v2x, v2y, v2z)*NORMA(v3x, v3y, v3z)));
  if (ang_max < ang) ang_max = ang;

  VetorTangente(&v1x, &v1y, &v1z, f1, f2, f3, u1, v1, 2);
  VetorTangente(&v2x, &v2y, &v2z, f1, f2, f3, umed, v1, 2);
  VetorTangente(&v3x, &v3y, &v3z, f1, f2, f3, u2, v1, 2);
  ang = acos((v1x*v2x + v1y*v2y + v1z*v2z)/
	(NORMA(v1x, v1y, v1z)*NORMA(v2x, v2y, v2z)));
  if (ang_max < ang) ang_max = ang;
  ang = acos((v2x*v3x + v2y*v3y + v2z*v3z)/
	(NORMA(v2x, v2y, v2z)*NORMA(v3x, v3y, v3z)));
  if (ang_max < ang) ang_max = ang;

  VetorTangente(&v1x, &v1y, &v1z, f1, f2, f3, u1, v2, 2);
  VetorTangente(&v2x, &v2y, &v2z, f1, f2, f3, umed, v2, 2);
  VetorTangente(&v3x, &v3y, &v3z, f1, f2, f3, u2, v2, 2);
  ang = acos((v1x*v2x + v1y*v2y + v1z*v2z)/
	(NORMA(v1x, v1y, v1z)*NORMA(v2x, v2y, v2z)));
  if (ang_max < ang) ang_max = ang;
  ang = acos((v2x*v3x + v2y*v3y + v2z*v3z)/
	(NORMA(v2x, v2y, v2z)*NORMA(v3x, v3y, v3z)));
  if (ang_max < ang) ang_max = ang;

  return ang_max;
}

/* ------------------------------------------------------------------------- */

float AnguloMax(float (*f1)(float, float), float (*f2)(float, float),
	     float (*f3)(float, float), float umin, float umax,
	     float vmin, float vmax, float inc_u, float inc_v,
	     float ang_saida) {

  float u, v, ang = -1000, angulo;

  for (u = umin; u < umax + inc_u/2; u += inc_u)
    for (v = vmin; v < vmax + inc_v/2; v += inc_v) {
      angulo = Angulo(f1, f2, f3, u, u + inc_u, v, v + inc_v);
      if (ang < angulo)	ang = angulo;
      if (ang > ang_saida) return ang;
    }
  return ang;
}

/* ------------------------------------------------------------------------- */

float BordasMax(float (*f1)(float, float), float (*f2)(float, float),
	     float (*f3)(float, float), float umin, float umax,
	     float vmin, float vmax, float inc_u, float inc_v,
	     float ang_saida) {

  float u, v, ang = -1000, angulo;

  for (u = umin; u < umax + inc_u/2; u += inc_u)
    for (v = vmin; v < vmax + inc_v/2; v += inc_v) {
      angulo = Bordas(f1, f2, f3, u, u + inc_u, v, v + inc_v);
      if (ang < angulo)	ang = angulo;
      if (ang > ang_saida) return ang;
    }
  return ang;
}

/* ------------------------------------------------------------------------- */

int ResolveSistema3x3(float A, float B, float C, float M,
		    float D, float E, float F, float N,
		    float G, float H, float I, float P,
		    float *X, float *Y, float *Z) {

  float Det;
  const float erro = 1E-10;

  Det = DET(A, B, C, D, E, F, G, H, I);
  if (fabs(Det) < erro) return 0;
  *X = DET(M, B, C, N, E, F, P, H, I)/Det;
  *Y = DET(A, M, C, D, N, F, G, P, I)/Det;
  *Z = DET(A, B, M, D, E, N, G, H, P)/Det;
  return 1;
}

/* ----------------------------------------------------------------------- */

int ResolveSistema3x4(float A, float B, float C, float D, float E,
		      float F, float G, float H, float I, float J,
		      float K, float L, float M, float N, float O,
		      float *X1, float *Y1, float *Z1, float *W1,
		      float *X2, float *Y2, float *Z2, float *W2) {

  float x2, y2, z2;

  if (!ResolveSistema3x3(A, B, C, E, F, G, H, J, K, L, M, O, X1, Y1, Z1))
    return 0;
  *W1 = 0;
  if (!ResolveSistema3x3(A, B, C, E - D, F, G, H, J - I, K, L, M, O - N,
    &x2, &y2, &z2)) return 0;
  *X2 = x2 - *X1;
  *Y2 = y2 - *Y1;
  *Z2 = z2 - *Z1;
  *W2 = 1;

  return 1;
}

/* ----------------------------------------------------------------------- */

#define ERRO 1e30

int IntersecaoDeTriangulos(float a1, float a2, float a3,
			   float b1, float b2, float b3,
			   float c1, float c2, float c3,
			   float A1, float A2, float A3,
			   float B1, float B2, float B3,
			   float C1, float C2, float C3,
			   float *p1, float *p2, float *p3,
			   float *q1, float *q2, float *q3) {

/* Determina a intersecao dos planos (a1, a2, a3) - (b1, b2, b3) - (c1, c2, c3)
   e (A1, A2, A3) - (B1, B2, B3) - (C1, C2, C3) */

  float sol, x1, y1, z1, w1, x2, y2, z2, w2, P1, P2, P3, Q1, Q2, Q3;
  int teste1, teste2, teste3, quant = 0;
  float t_aux, t1_1 = ERRO, t1_2 = ERRO, t2_1 = ERRO, t2_2 = ERRO;

  if (ResolveSistema3x4(b1 - a1, c1 - a1, A1 - B1, A1 - C1, A1 - a1,
			b2 - a2, c2 - a2, A2 - B2, A2 - C2, A2 - a2,
			b3 - a3, c3 - a3, A3 - B3, A3 - C3, A3 - a3,
			&x1, &y1, &z1, &w1, &x2, &y2, &z2, &w2)) {

    P1 = a1 + x1*(b1 - a1) + y1*(c1 - a1);
    P2 = a2 + x1*(b2 - a2) + y1*(c2 - a2);
    P3 = a3 + x1*(b3 - a3) + y1*(c3 - a3);

    Q1 = a1 + (x1 + x2)*(b1 - a1) + (y1 + y2)*(c1 - a1);
    Q2 = a2 + (x1 + x2)*(b2 - a2) + (y1 + y2)*(c2 - a2);
    Q3 = a3 + (x1 + x2)*(b3 - a3) + (y1 + y2)*(c3 - a3);

    teste1 = InterceptaSegmentos(P1, P2, P3, Q1, Q2, Q3,
				 a1, a2, a3, b1, b2, b3, &sol);
    if (teste1) {
      t1_1 = sol;
      quant++;
    }

    teste2 = InterceptaSegmentos(P1, P2, P3, Q1, Q2, Q3,
				 a1, a2, a3, c1, c2, c3, &sol);
    if (teste2) {
      if (quant)
	t1_2 = sol;
      else
	t1_1 = sol;
      quant++;
    }

    if (quant < 2) {
      teste3 = InterceptaSegmentos(P1, P2, P3, Q1, Q2, Q3,
				   b1, b2, b3, c1, c2, c3, &sol);
      if (teste3) {
	if (quant)
	  t1_2 = sol;
	else
	  t1_1 = sol;
	quant++;
      }
    }

    quant = 0;
    teste1 = InterceptaSegmentos(P1, P2, P3, Q1, Q2, Q3,
				 A1, A2, A3, B1, B2, B3, &sol);
    if (teste1) {
      t2_1 = sol;
      quant++;
    }

    teste2 = InterceptaSegmentos(P1, P2, P3, Q1, Q2, Q3,
				 A1, A2, A3, C1, C2, C3, &sol);
    if (teste2) {
      if (quant)
	t2_2 = sol;
      else
	t2_1 = sol;
      quant++;
    }

    if (quant < 2) {
      teste3 = InterceptaSegmentos(P1, P2, P3, Q1, Q2, Q3,
				   B1, B2, B3, C1, C2, C3, &sol);
      if (teste3) {
	if (quant)
	  t2_2 = sol;
	else
	  t2_1 = sol;
	quant++;
      }
    }

    if (t1_1 == ERRO || t1_2 == ERRO || t2_1 == ERRO || t2_2 == ERRO) return 0;

    if (t1_1 > t1_2) { t_aux = t1_1, t1_1 = t1_2, t1_2 = t_aux; }
    if (t2_1 > t2_2) { t_aux = t2_1, t2_1 = t2_2, t2_2 = t_aux; }

    if (-epsilon <= t1_1 && t1_1 <= t2_1 && t2_1 <= t2_2 && t2_2 <= t1_2 && t1_2 <= 1 + epsilon) {
      *p1 = P1 + t2_1*(Q1 - P1);
      *p2 = P2 + t2_1*(Q2 - P2);
      *p3 = P3 + t2_1*(Q3 - P3);
      *q1 = P1 + t2_2*(Q1 - P1);
      *q2 = P2 + t2_2*(Q2 - P2);
      *q3 = P3 + t2_2*(Q3 - P3);
      return 1;
    } else if (-epsilon <= t2_1 && t2_1 <= t1_1 && t1_1 <= t1_2 && t1_2 <= t2_2 && t2_2 <= 1 + epsilon) {
      *p1 = P1 + t1_1*(Q1 - P1);
      *p2 = P2 + t1_1*(Q2 - P2);
      *p3 = P3 + t1_1*(Q3 - P3);
      *q1 = P1 + t1_2*(Q1 - P1);
      *q2 = P2 + t1_2*(Q2 - P2);
      *q3 = P3 + t1_2*(Q3 - P3);
      return 1;
    } else if (-epsilon <= t1_1 && t1_1 <= t2_1 && t2_1 <= t1_2 && t1_2 <= t2_2 && t2_2 <= 1 + epsilon) {
      *p1 = P1 + t2_1*(Q1 - P1);
      *p2 = P2 + t2_1*(Q2 - P2);
      *p3 = P3 + t2_1*(Q3 - P3);
      *q1 = P1 + t1_2*(Q1 - P1);
      *q2 = P2 + t1_2*(Q2 - P2);
      *q3 = P3 + t1_2*(Q3 - P3);
      return 1;
    } else if (-epsilon <= t2_1 && t2_1 <= t1_1 && t1_1 <= t2_2 && t2_2 <= t1_2 && t1_2 <= 1 + epsilon) {
      *p1 = P1 + t1_1*(Q1 - P1);
      *p2 = P2 + t1_1*(Q2 - P2);
      *p3 = P3 + t1_1*(Q3 - P3);
      *q1 = P1 + t2_2*(Q1 - P1);
      *q2 = P2 + t2_2*(Q2 - P2);
      *q3 = P3 + t2_2*(Q3 - P3);
      return 1;
    }
  }

  return 0;
}

/* ----------------------------------------------------------------------- */

int QuantasSubdivisoes(float (*f1)(float, float), float (*f2)(float, float),
	     float (*f3)(float, float), float umin, float umax, float vmin,
	     float vmax, float normal, float bordas, int subdiv_max) {

/* Determina a quantidades de subdivisoes da regiao [umin, umax] x [vmin, vmax]
   de forma a  que os  angulos entre os vetores normais nao ultrapassem o valor
   normal  em  graus e  que  os vetores  tangentes nas bordas nao ultrapassem o
   valor bordas em graus */

  unsigned int n_subdiv = 0;
  float bordas_max = 1000, angulo_max = 1000, inc_u, inc_v,
	angulo_bordas_max, angulo_normal_max;

  angulo_normal_max = normal*M_PI/180.0;
  angulo_bordas_max = bordas*M_PI/180.0;

  do {
    n_subdiv += 10;
    inc_u = (umax - umin)/n_subdiv;
    inc_v = (vmax - vmin)/n_subdiv;
    if (angulo_max > angulo_normal_max)
      angulo_max = AnguloMax(f1, f2, f3, umin, umax, vmin, vmax,
			   inc_u, inc_v, angulo_normal_max);
    if (bordas_max > angulo_bordas_max)
      bordas_max = BordasMax(f1, f2, f3, umin, umax, vmin, vmax,
			   inc_u, inc_v, angulo_bordas_max);
  } while (n_subdiv <= subdiv_max && (angulo_max > angulo_normal_max ||
				      bordas_max > angulo_bordas_max));

  return n_subdiv;
}

/* ----------------------------------------------------------------------- */

#define TRISUP(f1,f2,f3,g1,g2,g3,u1,v1,u2,v2,u3,v3,p1,p2,p3,q1,q2,q3) \
  IntersecaoDeTriangulos(f1(u1, v1), f2(u1, v1), \
  f3(u1, v1), f1(u2, v2), f2(u2, v2), f3(u2, v2), f1(u3, v3), f2(u3, v3), \
  f3(u3, v3), g1(u1, v1), g2(u1, v1), g3(u1, v1), g1(u2, v2), g2(u2, v2), \
  g3(u2, v2), g1(u3, v3), g2(u3, v3), g3(u3, v3), &p1, &p2, &p3, &q1, &q2, &q3)

int TriSup(float (*f1)(float, float), float (*f2)(float, float),
			     float (*f3)(float, float),
			     float (*g1)(float, float),
			     float (*g2)(float, float),
			     float (*g3)(float, float),
			     float u1, float v1,
			     float u2, float v2,
			     float u3, float v3,
			     float *p1, float *p2, float *p3,
			     float *q1, float *q2, float *q3) {

  return IntersecaoDeTriangulos(f1(u1, v1), f2(u1, v1), f3(u1, v1),
				f1(u2, v2), f2(u2, v2), f3(u2, v2),
				f1(u3, v3), f2(u3, v3), f3(u3, v3),
				g1(u1, v1), g2(u1, v1), g3(u1, v1),
				g1(u2, v2), g2(u2, v2), g3(u2, v2),
				g1(u3, v3), g2(u3, v3), g3(u3, v3),
				p1, p2, p3,
				q1, q2, q3);
}

/* ----------------------------------------------------------------------- */

#define YELLOW 0
unsigned long quant_pontos = 1;

#define MAX_N 1

int DesenhaReta(float a1, float a2, float a3,
		 float b1, float b2, float b3, int cor) {

/* Desenha a reta ligando os pontos (a1, a2, a3) - (b1, b2, b3) */
   static int N = 0;

   printf("\n\nSegmento de reta obtido ligando (%f, %f, %f) a (%f, %f, %f)",
	    a1, a2, a3, b1, b2, b3);

   if (++N > MAX_N) return 0;
   return 1;
}

/* ----------------------------------------------------------------------- */

void Tec3(float (*f1)(float, float), float (*f2)(float, float),
			     float (*f3)(float, float),
			     float (*g1)(float, float),
			     float (*g2)(float, float),
			     float (*g3)(float, float),
			     float umin, float umax,
			     float vmin, float vmax, int n_subdiv) {

/* Implementa a tecnica 3: subdivisao dos dominios */

  int quant, teste, N, M, n, m;
  float u, v, U, V, inc_U, inc_V, inc_u, inc_v,
	p1, p2, p3, q1, q2, q3, P1, P2, P3, Q1, Q2, Q3;

  quant = (int) quant_pontos;

  inc_u = (umax - umin)/n_subdiv;
  inc_v = (vmax - vmin)/n_subdiv;
  inc_U = inc_u/quant;
  inc_V = inc_v/quant;

  for (N = 0; N < n_subdiv; N++)
    for (M = 0; M < n_subdiv; M++)
      for (u = umin + N*inc_u; u < umin + (N + 1)*inc_u; u += inc_u)
	for (v = vmin + M*inc_v; v < vmin + (M + 1)*inc_v; v += inc_v) {
/*
(u, v + inc_v) +----------+ (u + inc_u, v + inc_v)
	       |	  |
	       +----------+
	(u, v)             (u + inc_u, v)
*/
      teste = TriSup(f1, f2, f3, g1, g2, g3,
	      u, v, u + inc_u, v, u + inc_u, v + inc_v,
	      &p1, &p2, &p3, &q1, &q2, &q3) ||
	      TriSup(f1, f2, f3, g1, g2, g3,
	      u, v, u, v + inc_v, u + inc_u, v + inc_v,
	      &P1, &P2, &P3, &Q1, &Q2, &Q3) ||
	      TriSup(f1, f2, f3, g1, g2, g3,
	      u + inc_u, v, u, v + inc_v, u + inc_u, v + inc_v,
	      &P1, &P2, &P3, &Q1, &Q2, &Q3) ||
	      TriSup(f1, f2, f3, g1, g2, g3,
	      u + inc_u, v, u, v + inc_v, u, v,
	      &P1, &P2, &P3, &Q1, &Q2, &Q3);

      if (teste) {

	for (n = 0; n < quant; n++)
	  for (m = 0; m < quant; m++)
	    for (U = u + n*inc_U; U <= u + (n + 1)*inc_U; U += inc_U)
	      for (V = v + m*inc_V; V <= v + (m + 1)*inc_V; V += inc_V) {

	    teste = TriSup(f1, f2, f3, g1, g2, g3,
	      U, V, U + inc_U, V, U + inc_U, V + inc_V,
	      &p1, &p2, &p3, &q1, &q2, &q3);
	    if (teste) { if (!DesenhaReta(p1, p2, p3, q1, q2, q3, YELLOW)) return;
				 }
	    teste = TriSup(f1, f2, f3, g1, g2, g3,
	      U, V, U, V + inc_V, U + inc_U, V + inc_V,
	      &P1, &P2, &P3, &Q1, &Q2, &Q3);

	    if (teste) { if (!DesenhaReta(P1, P2, P3, Q1, Q2, Q3, YELLOW)) return;
			 }
	    teste = TriSup(f1, f2, f3, g1, g2, g3,
	      U + inc_U, V, U, V + inc_V, U + inc_U, V + inc_V,
	      &P1, &P2, &P3, &Q1, &Q2, &Q3);
	    if (teste) { if (!DesenhaReta(P1, P2, P3, Q1, Q2, Q3, YELLOW)) return;
			 }
	    teste = TriSup(f1, f2, f3, g1, g2, g3,
	      U + inc_U, V, U, V + inc_V, U, V,
	      &P1, &P2, &P3, &Q1, &Q2, &Q3);
	    if (teste) { if (!DesenhaReta(P1, P2, P3, Q1, Q2, Q3, YELLOW)) return;
		 }
	  }

      }  /* fim do if (teste) ... */
    } /* fim do for (v ... */
  }

/*** FIM DE DIV.C ***/
